-- Sequence of unique values for request ID numbers.
-- A request ID number is generated when a new request arrives.
-- A request record is created with this ID.
-- Subsequent processing refers to this ID.

DROP SEQUENCE request_id_sequence IF EXISTS;
COMMIT;

CREATE SEQUENCE request_id_sequence;
COMMIT;


-- A table to store the database version number.
-- Prevents accidentally damaging an incompatible database.

DROP TABLE db_version IF EXISTS;
COMMIT;

CREATE MEMORY TABLE db_version (
version INT
);
COMMIT;

INSERT INTO db_version (version) VALUES ('1');
COMMIT;


-- A table to enumerate request types (among them recognized errors).
-- Must match internal constants of class "Worker".
-- ID values are manual to avoid mistakes.

DROP TABLE request_type IF EXISTS;
COMMIT;

CREATE MEMORY TABLE request_type (
id INT PRIMARY KEY,
name VARCHAR(64)
);
COMMIT;

INSERT INTO request_type (id,name) VALUES (0,'Failed opening comms.');
INSERT INTO request_type (id,name) VALUES (1,'Failed reading request.');

INSERT INTO request_type (id,name) VALUES (2,'Denied by total quota.');

INSERT INTO request_type (id,name) VALUES (3,'Failed extracting line.');
INSERT INTO request_type (id,name) VALUES (4,'Failed URL decoding line.');
INSERT INTO request_type (id,name) VALUES (5,'Failed classifying.');

INSERT INTO request_type (id,name) VALUES (6,'Denied by PUT quota.');

INSERT INTO request_type (id,name) VALUES (7,'Failed parsing.');
INSERT INTO request_type (id,name) VALUES (8,'Failed composing response.');

INSERT INTO request_type (id,name) VALUES (9,'Get destinations.');
INSERT INTO request_type (id,name) VALUES (10,'Get webcaches.');
INSERT INTO request_type (id,name) VALUES (11,'Put destination.');
INSERT INTO request_type (id,name) VALUES (12,'Put webcache.');
INSERT INTO request_type (id,name) VALUES (13,'Put both.');

INSERT INTO request_type (id,name) VALUES (14,'Browse index.');
INSERT INTO request_type (id,name) VALUES (15,'Browse statistics.');
COMMIT;


-- A table to enumerate response types (among them not answering).
-- Must match internal constants of class "Worker".
-- ID values are manual to avoid mistakes.

DROP TABLE response_type IF EXISTS;
COMMIT;

CREATE MEMORY TABLE response_type (
id INT PRIMARY KEY,
name VARCHAR(64)
);
COMMIT;

INSERT INTO response_type (id,name) VALUES (0,'Dropped.');
INSERT INTO response_type (id,name) VALUES (1,'Error, sent error message.');
INSERT INTO response_type (id,name) VALUES (2,'Rejected, normal response.');
INSERT INTO response_type (id,name) VALUES (3,'Accepted, normal response.');
COMMIT;


-- A table to store recent requests.
-- ID values are manual since we use sequence to generate them.

DROP TABLE request IF EXISTS;
COMMIT;

CREATE CACHED TABLE request (
id BIGINT PRIMARY KEY,
arrived TIMESTAMP,
request_type_id INT,
response_type_id INT,
from_destination VARCHAR(1024),
put_destination VARCHAR(1024),
put_webcache VARCHAR(1024),
headers VARCHAR(2048),
abusive BOOLEAN
);
COMMIT;


-- Indexes to speed up searches
-- against various interesting request parameters.

CREATE INDEX request_arrived ON request(arrived);
COMMIT;

CREATE INDEX request_request_type_id ON request(request_type_id);
COMMIT;

CREATE INDEX request_response_type_id ON request(response_type_id);
COMMIT;

CREATE INDEX request_from_destination ON request(from_destination);
COMMIT;

CREATE INDEX request_put_destnation ON request(put_destination);
COMMIT;

