package jwebcache;

import org.apache.log4j.Logger;

import java.util.*;
import java.io.*;
import java.sql.*;


public class Database extends Thread {

    // Our logger instance.
    private static Logger logger = Logger.getLogger(Database.class);
    
    // Configuration.
    private Properties properties; // Our config properties.

    private int returnDestinations; // How many destinations to return per query.
    private int returnWebcaches; // How many webcaches to return per query.
    private int maintenanceIntervalMs; // How often to perform DB maintenance.
    private int maxRecords; // How many records may be kept?

    private int totalQuota; // How many requests total may be submitted?
    private int putQuota; // How many PUT requests may be submitted?
    private int totalQuotaMinutes; // During how many minutes?
    private int putQuotaMinutes; // During how many minutes?

    // References to important objects.
    private Connection conn; // Connection to the database server.

    // Compatibility.
    private final int COMPATIBLE_DB_VERSION = 1; // DB version we support.
    
    // Shutdown flag, raised to inform that run() should exit
    // and database maintenance not occur anymore.
    // Access to this variable should be synchronized.
    private Boolean shuttingDown = false;


    public Database(Properties properties) throws Exception {

        // Grab arguments to local variables
        this.properties = properties;
        
        try {
            // Parse configuration properties
            returnDestinations = Integer.parseInt(properties.getProperty("Database.returnDestinations","20"));
            returnWebcaches = Integer.parseInt(properties.getProperty("Database.returnWebcaches","10"));
            maintenanceIntervalMs = Integer.parseInt(properties.getProperty("Database.maintenanceIntervalMs","180000"));
            maxRecords = Integer.parseInt(properties.getProperty("Database.maxRecords","5000"));
            totalQuota = Integer.parseInt(properties.getProperty("Database.totalQuota","15"));
            putQuota = Integer.parseInt(properties.getProperty("Database.putQuota","3"));
            totalQuotaMinutes = Integer.parseInt(properties.getProperty("Database.totalQuotaMinutes","60"));
            putQuotaMinutes = Integer.parseInt(properties.getProperty("Database.putQuotaMinutes","60"));
        } catch (Exception e) {
            logger.fatal("Failed reading config properties, database constructor aborting.",e);
            shutdown();
            throw new Exception("Database constructor aborting.");
        }

        try {
            logger.info("Loading JDBC driver.");
            Class.forName("org.hsqldb.jdbcDriver");
        } catch (Exception e) {
            logger.fatal("Failed loading JDBC driver, database constructor aborting.",e);
            shutdown();
            throw new Exception("Database constructor aborting.");
        }

        String location = properties.getProperty("Database.location");
        if (location == null)
            location = "database/db";
        else
            location += "/database/db";
        try {
            logger.info("Opening JDBC connection.");
            conn = DriverManager.getConnection("jdbc:hsqldb:file:" + location,"sa","");
        } catch (Exception e) {
            logger.fatal("Failed opening JDBC connection, database constructor aborting.",e);
            shutdown();
            throw new Exception("Database constructor aborting.");
        }
        
        // Current HSQLDB does not seem to auto-commit on shutdown,
        // and messing with write delays under such conditions ain't my idea of fun.
        logger.info("Disabling write delay.");
        
        if (!executeStatement("SET WRITE_DELAY FALSE")) {
            logger.fatal("Failed disabling write delay, database constructor aborting.");
            shutdown();
            throw new Exception("Database constructor aborting.");
        }
        
        logger.info("Searching/upgrading for compatible database version " + COMPATIBLE_DB_VERSION + ".");

        if (!makeCompatibleDbVersion()) {
            logger.fatal("Failed reaching compatible database version, database constructor aborting.");
            shutdown();
            throw new Exception("Database constructor aborting.");
        }
    }

    
    public void run() {

        while (true) {
            
            // Secure exclusive access to shutdown status.
            synchronized(shuttingDown) {

                // Verify that shutdown hasn't run.
                if (shuttingDown) {
                    logger.info("Database maintenance quitting, since shutdown has started.");
                    return;
                }
                
                logger.info("Performing database maintenance.");
                purgeOldRecords();
            }

            // Out of the synchronized block, idle for a long while.
            try { sleep(maintenanceIntervalMs); } catch (Exception e) { }
        }
    }
    

    public void shutdown() {

        // Secure exclusive access to shutdown status.
        synchronized(shuttingDown) {
            
            // Announce that shutdown has run.
            shuttingDown = true;
        }

        logger.info("Compacting database and shutting down.");
        shutdownCompact();

        try {
            logger.info("Closing JDBC connection.");
            conn.close();
        } catch (SQLException e) {
            logger.error("Failed closing JDBC connection!",e);
        } finally {
            logger.debug("Nulling JDBC connection.");
            conn = null;
        }
    }
    

    private boolean makeCompatibleDbVersion() {

        // Get database version, check against errors
        int version = getDbVersion();
        if (version < 0) {
            logger.fatal("Unknown database version, incompatible!");
            return false;
        }
        
        // Upgrade as necessary
        while (version < COMPATIBLE_DB_VERSION) {
            
            int nextVersion = version + 1;
            logger.info("Upgrading database to version " + nextVersion + ".");
            
            if (upgradeToDbVersion(nextVersion)) {
                logger.info("Upgraded database to version " + nextVersion + ".");
                version = nextVersion;
            } else {
                logger.fatal("Failed upgrading database to version " + nextVersion + "!");
                break;
            }
        }
        
        // Check which version we finally reached
        if (version == COMPATIBLE_DB_VERSION) {
            logger.info("Final database version " + version + " is compatible.");
            return true;
        } else {
            logger.fatal("Final database version " + version + " is incompatible!");
            return false;
        }
    }

    
    private int getDbVersion() {
        
        int version = -1;
        PreparedStatement st = null;
        ResultSet rs = null;
        
        try {
            logger.debug("Preparing to read database version.");
            st = conn.prepareStatement("SELECT version FROM db_version");
            
            logger.debug("Executing query.");
            rs = st.executeQuery();
            
            if (rs.next()) {
                version = rs.getInt("version");
                logger.info("Found database version " + version + ".");
            }
            else throw new Exception("Version table exists but empty.");
            
        } catch (SQLException e) {
            
            // Ugly but useful code: check SQL result codes
            // to recognize a "table does not exist" condition (new install).
            if ((e.getErrorCode() == -22) && (e.getSQLState().equals("S0002"))) {
                logger.info("Version table not found, recognizing as version 0.");
                version = 0;
            }
            else logger.fatal("Failed getting DB version, unforeseen SQL exception.",e);
            
        } catch (Exception e) {
            logger.fatal("Failed getting DB version, unforeseen exception.",e);
        }
        
        return version;
    }

    

    private boolean upgradeToDbVersion(int version) {
        
        boolean success = false;
        FileInputStream fis = null;
        BufferedReader reader = null;
        StringBuffer buffer = new StringBuffer();
        String line = null;
        
        String location = properties.getProperty("Database.location");
        if (location == null)
            location = System.getProperty("user.dir");
        File script = new File(location, "config/dbversion" + version + ".sql");

        try {
            logger.debug("Opening upgrade script for version " + version + ".");
            fis = new FileInputStream(script);
            reader = new BufferedReader(new InputStreamReader(fis));
            
            // Read lines while they aren't null
            while ((line = reader.readLine()) != null) {
                    
                // Trim whitespace, potentially log.
                line = line.trim();
                logger.debug(line);
                    
                // Skip comments, collect lines into buffer.
                if (line.startsWith("--")) continue;
                buffer.append(line);
                    
                // If line finishes a command.
                if (line.endsWith(";")) {
                    
                    // Execute buffer, then clear it.
                    if (!executeStatement(buffer.toString())) {
                        throw new Exception("Failed executing upgrade script statement.");
                    }
                    buffer.setLength(0);
                }
            }

            // Upgrade was successful
            success = true;
            
        } catch (IOException e) {
            logger.fatal("Failed reading upgrade script.",e);
        } catch (Exception e) {
            logger.fatal("Failed running upgrade script.",e);
        } finally {
            if (reader != null) try { reader.close(); } catch (IOException ioe) {}
            if (fis != null) try { fis.close(); } catch (IOException ioe) {}
        }
        
        return success;
    }

    
    private void cleanStatement(PreparedStatement st) {
        
        if (st != null) {
            try {
                logger.debug("Closing statement.");
                st.close();
            } catch (SQLException e) {
                 logger.error("Failed closing statement, nulling anyway.",e);
            } finally {
                st = null;
            }
        } else logger.warn("Statement already null.");
    }
    
    
    private void cleanResultSet(ResultSet rs) {
        
        if (rs != null) {
            try {
                logger.debug("Closing result set.");
                rs.close();
            } catch (SQLException e) {
                logger.error("Failed closing result set, nulling anyway.",e);
            } finally {
                rs = null;
            }
        } else logger.warn("Result set already null.");
    }

    
    private boolean executeStatement(String statement) {

        boolean success = false;
        PreparedStatement st = null;

        try {
            logger.debug("Preparing statement.");
            st = conn.prepareStatement(statement);

            logger.debug("Executing statement.");
            st.executeUpdate();
            success = true;

        } catch (SQLException e) {
            logger.error("Failed executing statement!",e);
        }
        cleanStatement(st);

        return success;
    }

    
    private final String SQL_SHUTDOWN_COMPACT =
    "SHUTDOWN COMPACT";
    
    private boolean shutdownCompact() {
        
        boolean success = false;
        PreparedStatement st = null;

        try {
            logger.debug("Preparing to shutdown and compact database.");
            st = conn.prepareStatement(SQL_SHUTDOWN_COMPACT);
            
            logger.debug("Executing statement.");
            st.execute();
            success = true;

        } catch (SQLException e) {
            logger.error("Failed shutdown or compacting of database!",e);
        }
        cleanStatement(st);

        return success;
    }

    
    private final String SQL_PURGE_OLD_RECORDS =
    "DELETE FROM request WHERE id < (SELECT (MIN(id) + COUNT(id) - ?) FROM request)";
    
    private boolean purgeOldRecords() {
        
        boolean success = false;
        PreparedStatement st = null;

        try {
            logger.debug("Preparing to purge old records.");
            st = conn.prepareStatement(SQL_PURGE_OLD_RECORDS);
            
            logger.debug("Attaching values and executing statement.");
            st.setLong(1, maxRecords);
            int recordsAffected = st.executeUpdate();
            logger.info("Purged " + recordsAffected + " old records.");
            success = true;

        } catch (SQLException e) {
            logger.error("Failed purging old records!",e);
        }
        cleanStatement(st);
        
        return success;
    }

    
    private final String SQL_GET_REQUEST_ID =
    "CALL NEXT VALUE FOR request_id_sequence";
    
    public long getRequestId() {

        long requestId = -1;
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.debug("Preparing to get next request ID.");
            st = conn.prepareStatement(SQL_GET_REQUEST_ID);
            
            logger.debug("Executing statement.");
            rs = st.executeQuery();
            if (rs.next()) requestId = rs.getLong(1);

        } catch (SQLException e) {
            logger.error("Failed getting next request ID!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);
        
        return requestId;
    }


    private final String SQL_REGISTER_REQUEST =
    "INSERT INTO request (id,arrived,from_destination) VALUES (?, NOW(), ?)";
    
    public boolean registerRequest(long requestId, String fromDestination) {
        
        boolean success = false;
        PreparedStatement st = null;

        try {
            logger.debug("Preparing to register request " + requestId + ".");
            st = conn.prepareStatement(SQL_REGISTER_REQUEST);
            
            logger.debug("Attaching values and executing statement.");
            st.setLong(1, requestId);
            st.setString(2, fromDestination);
            st.executeUpdate();
            success = true;

        } catch (SQLException e) {
            logger.error("Failed registering request!",e);
        }
        cleanStatement(st);
        
        return success;
    }
    
    
    private final String SQL_LOG_HEADERS =
    "UPDATE request SET headers=? WHERE (id=?)";
    
    public boolean logHeaders(long requestId, String headers) {
        
        boolean success = false;
        PreparedStatement st = null;

        try {
            logger.debug("Updating request " + requestId + " with headers.");
            st = conn.prepareStatement(SQL_LOG_HEADERS);
            
            logger.debug("Attaching values and executing statement.");
            st.setString(1,headers);
            st.setLong(2,requestId);
            st.executeUpdate();
            success = true;

        } catch (SQLException e) {
            logger.error("Failed logging headers!",e);
        }
        cleanStatement(st);
        
        return success;
    }

    
    private final String SQL_UPDATE_REQUEST =
    "UPDATE request SET request_type_id=?, response_type_id=? WHERE (id=?)";
    
    public boolean updateRequest(long requestId, int requestType, int responseType) {
        
        boolean success = false;
        PreparedStatement st = null;

        try {
            logger.info("Updating " + requestId + " to request type " + requestType + ", response type " + responseType + ".");
            st = conn.prepareStatement(SQL_UPDATE_REQUEST);
            
            logger.debug("Attaching values and executing statement.");
            st.setInt(1,requestType);
            st.setInt(2,responseType);
            st.setLong(3,requestId);
            st.executeUpdate();
            success = true;

        } catch (SQLException e) {
            logger.error("Failed updating request!",e);
        }
        cleanStatement(st);
        
        return success;
    }

    
    private final String SQL_PUT_DESTINATION =
    "UPDATE request SET put_destination=? WHERE (id=?)";
    
    public boolean putDestination(long requestId, String destination) {
        
        boolean success = false;
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.info("Updating request " + requestId + " with destination " + destination + ".");
            st = conn.prepareStatement(SQL_PUT_DESTINATION);
            
            logger.debug("Attaching values and executing statement.");
            st.setString(1,destination);
            st.setLong(2,requestId);
            st.executeUpdate();
            success = true;

        } catch (SQLException e) {
            logger.error("Failed updating put destination!",e);
        }
        cleanStatement(st);
        
        return success;
    }
    
    
    private final String SQL_PUT_WEBCACHE =
    "UPDATE request SET put_webcache=? WHERE (id=?)";
    
    public boolean putWebcache(long requestId, String url) {
        
        boolean success = false;
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.info("Updating request " + requestId + " with URL " + url + ".");
            st = conn.prepareStatement(SQL_PUT_WEBCACHE);
            
            logger.debug("Attaching values and executing statement.");
            st.setString(1,url);
            st.setLong(2,requestId);
            st.executeUpdate();
            success = true;

        } catch (SQLException e) {
            logger.error("Failed updating put destination!",e);
        }
        cleanStatement(st);
        
        return success;
    }


    private final String SQL_GET_TOTAL_COUNT =
    "SELECT COALESCE(COUNT(from_destination),0) other_count FROM request " +
    "WHERE (DATEDIFF('minute',request.arrived,NOW()) < ?) AND " +
    "(from_destination=?) GROUP BY from_destination";
    
    public boolean exceedsTotalQuota(String fromDestination) {

        boolean exceeds = true; // The worse assumption is "true" here.
        PreparedStatement st = null;
        ResultSet rs = null;
        int count = 0;

        try {
            logger.debug("Preparing to get total request count.");
            st = conn.prepareStatement(SQL_GET_TOTAL_COUNT);
            
            logger.debug("Attaching values and executing statement.");
            st.setInt(1,totalQuotaMinutes);
            st.setString(2,fromDestination);
            rs = st.executeQuery();
            
            if (rs.next()) count = rs.getInt(1); else count = 0;
            if (count < totalQuota) {
                logger.debug("Total request count " + count + " fits quota " + totalQuota + ".");
                exceeds = false;
            }
            else logger.debug("Total request count " + count + " exceeds quota " + totalQuota + ".");
            
        } catch (SQLException e) {
            logger.error("Failed getting total request count!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        return exceeds;
    }
    
    
    private final String SQL_GET_PUT_COUNT =
    "SELECT COALESCE(COUNT(from_destination),0) put_count FROM request " +
    "WHERE (put_destination IS NOT NULL) AND (DATEDIFF('minute',request.arrived,NOW()) < ?) AND " +
    "(from_destination=?) GROUP BY from_destination";
    
    public boolean exceedsPutQuota(String fromDestination) {

        boolean exceeds = true; // The worse assumption is "true" here.
        PreparedStatement st = null;
        ResultSet rs = null;
        int count = 0;

        try {
            logger.debug("Preparing to get PUT request count.");
            st = conn.prepareStatement(SQL_GET_PUT_COUNT);
            
            logger.debug("Attaching values and executing statement.");
            st.setInt(1,putQuotaMinutes);
            st.setString(2,fromDestination);
            rs = st.executeQuery();
            
            if (rs.next()) count = rs.getInt(1); else count = 0;
            if (count < putQuota) {
                logger.debug("PUT request count " + count + " fits quota " + putQuota + ".");
                exceeds = false;
            }
            else logger.debug("PUT request count " + count + " exceeds quota " + putQuota + ".");
            
        } catch (SQLException e) {
            logger.error("Failed getting PUT request count!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        return exceeds;
    }

    
    private final String SQL_GET_DESTINATIONS = 
    "SELECT TOP ? DISTINCT put_destination,MAX(id) id FROM request " +
    "WHERE (put_destination IS NOT NULL) GROUP BY put_destination ORDER BY id DESC";
            
    public String getDestinations() {

        StringBuffer result = new StringBuffer("");
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.debug("Preparing to get destinations.");
            st = conn.prepareStatement(SQL_GET_DESTINATIONS);
            
            logger.debug("Attaching values and executing statement.");
            st.setInt(1,returnDestinations);
            rs = st.executeQuery();
            while (rs.next()) {
                result.append(rs.getString(1));
                result.append("\r\n");
            }

        } catch (SQLException e) {
            logger.error("Failed getting destinations!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);
        
        return result.toString();
    }
    
    
    private final String SQL_GET_WEBCACHES =
    "SELECT TOP ? DISTINCT put_webcache,MAX(id) id FROM request " +
    "WHERE (put_webcache IS NOT NULL) GROUP BY put_webcache ORDER BY id DESC";

    public String getWebcaches() {

        StringBuffer result = new StringBuffer("");
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.debug("Preparing to get webcaches.");
            st = conn.prepareStatement(SQL_GET_WEBCACHES);
            
            logger.debug("Attaching values and executing statement.");
            st.setInt(1,returnWebcaches);
            rs = st.executeQuery();
            while (rs.next()) {
                result.append(rs.getString(1));
                result.append("\r\n");
            }

        } catch (SQLException e) {
            logger.error("Failed getting webcaches!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        return result.toString();
    }
    
    
    private final String SQL_GET_RECENT_REQUESTS =
    "SELECT request.id, INSERT(LEFT(request.arrived,19),12,2,'xx') arrived, LEFT(request.from_destination,4) from_destination, " +
    "COALESCE(request_type.name,'(still processing)') request_type, " +
    "COALESCE(response_type.name,'(still processing)') response_type " +
    "FROM request " +
    "LEFT JOIN request_type ON request.request_type_id=request_type.id " +
    "LEFT JOIN response_type ON request.response_type_id=response_type.id " +
    "ORDER BY id DESC LIMIT 20";
    
    public String getRecentRequests() {

        StringBuffer result = new StringBuffer("");
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.debug("Preparing to get recent requests.");
            st = conn.prepareStatement(SQL_GET_RECENT_REQUESTS);
            
            logger.debug("Executing statement.");
            rs = st.executeQuery();
            
            while (rs.next()) {
                result.append("<tr><td>");
                result.append(rs.getString(1));
                result.append("</td><td>");
                result.append(rs.getString(2));
                result.append("</td><td>");
                result.append(rs.getString(3));
                result.append("</td><td>");
                result.append(rs.getString(4));
                result.append("</td><td>");
                result.append(rs.getString(5));
                result.append("</td></tr>\n");
            }

        } catch (SQLException e) {
            logger.error("Failed getting recent requests!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        return result.toString();
    }
    
    
    private final String SQL_GET_RECENT_HEADERS =
    "SELECT request.id, INSERT(LEFT(request.arrived,19),12,2,'xx') arrived, " +
    "LEFT(request.from_destination,4) from_destination, " + 
    "request_type.name request_type, response_type.name response_type, " +
    "COALESCE(request.headers,'Not available. ') headers FROM request " +
    "LEFT JOIN request_type ON request.request_type_id=request_type.id " +
    "LEFT JOIN response_type ON request.response_type_id=response_type.id " +
    "ORDER BY id DESC LIMIT 5";
    
    public String getRecentHeaders() {

        StringBuffer result = new StringBuffer("");
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.debug("Preparing to get recent headers.");
            st = conn.prepareStatement(SQL_GET_RECENT_HEADERS);
            
            logger.debug("Executing statement.");
            rs = st.executeQuery();
            
            while (rs.next()) {
                result.append("<tr><td><div class=headers1>");
                result.append(rs.getString(1));
                result.append("<br>");
                result.append(rs.getString(2));
                result.append("<br>");
                result.append(rs.getString(3));
                result.append("<br>");
                result.append(rs.getString(4));
                result.append("<br>");
                result.append(rs.getString(5));
                result.append("</div></td><td><div class=headers2>");
                result.append(rs.getString(6));
                result.append("</div></td></tr>\n");
            }

        } catch (SQLException e) {
            logger.error("Failed getting recent headers!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        return result.toString();
    }
    
    
    private final String SQL_GET_PUT_QUOTA_SUMMARY =
    "SELECT TOP 10 LEFT(from_destination,4) from_destination, COUNT(from_destination) put_count " +
    "FROM request WHERE (put_destination IS NOT NULL) AND " +
    "(DATEDIFF('minute',request.arrived,NOW()) < ?) " +
    "GROUP BY from_destination ORDER BY put_count DESC";
    
    public String getPutQuotaSummary() {

        StringBuffer result = new StringBuffer("");
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.debug("Preparing to get PUT quota summary.");
            st = conn.prepareStatement(SQL_GET_PUT_QUOTA_SUMMARY);
            
            logger.debug("Attaching values and executing statement.");
            st.setInt(1,putQuotaMinutes);
            rs = st.executeQuery();
            
            while (rs.next()) {
                result.append("<tr><td class=dark>");
                result.append(rs.getString(1));
                result.append("</td><td class=light>");
                result.append(rs.getString(2));
                result.append("</td></tr>\n");
            }

        } catch (SQLException e) {
            logger.error("Failed getting PUT quota summary!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        return result.toString();
    }
    
    
    private final String SQL_GET_TOTAL_QUOTA_SUMMARY =
    "SELECT TOP 10 LEFT(from_destination,4) from_destination, COUNT(from_destination) other_count " +
    "FROM request WHERE (put_destination IS NULL) AND " +
    "(DATEDIFF('minute',request.arrived,NOW()) < ?) " +
    "GROUP BY from_destination ORDER BY other_count DESC";
    
    public String getTotalQuotaSummary() {

        StringBuffer result = new StringBuffer("");
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.debug("Preparing to get total quota summary.");
            st = conn.prepareStatement(SQL_GET_TOTAL_QUOTA_SUMMARY);
            
            logger.debug("Attaching values and executing statement.");
            st.setInt(1,totalQuotaMinutes);
            rs = st.executeQuery();
            
            while (rs.next()) {
                result.append("<tr><td class=dark>");
                result.append(rs.getString(1));
                result.append("</td><td class=light>");
                result.append(rs.getString(2));
                result.append("</td></tr>\n");
            }

        } catch (SQLException e) {
            logger.error("Failed getting total quota summary!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        return result.toString();
    }

    
    private final String SQL_GET_REQUEST_TYPE_NAMES =
    "SELECT name FROM request_type";
    
    public Vector getRequestTypeNames() {
        
        Vector<String> result = new Vector<String>();
        PreparedStatement st = null;
        ResultSet rs = null;
                
        try {
            logger.debug("Preparing to get request type names.");
            st = conn.prepareStatement(SQL_GET_REQUEST_TYPE_NAMES);
            
            logger.debug("Attaching values and executing statement.");
            rs = st.executeQuery();
            
            while (rs.next()) {
                result.add(rs.getString(1));
            }

        } catch (SQLException e) {
            logger.error("Failed getting request type names!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        return result;
    }
    
    
    private final String SQL_GET_REQUEST_TYPE_TOTALS =
    "SELECT request_type.id, COALESCE(COUNT(period_data.request_type_id),0) subtotal " +
    "FROM request_type LEFT JOIN " +
    "(SELECT request.request_type_id FROM request " +
    "WHERE (DATEDIFF(?,request.arrived,NOW()) >= ?) AND (DATEDIFF(?,request.arrived,NOW()) <= ?)) period_data " +
    "ON (request_type.id=period_data.request_type_id) " +
    "GROUP BY id";
    
    private int[] getRequestTypeTotals(String timeUnit, int minValue, int maxValue) {
        
        int[] result;
        Vector<Integer> temp = new Vector<Integer>();
        PreparedStatement st = null;
        ResultSet rs = null;

        try {
            logger.debug("Preparing to get request type counts.");
            st = conn.prepareStatement(SQL_GET_REQUEST_TYPE_TOTALS);
            
            logger.debug("Attaching values and executing statement.");
            st.setString(1,timeUnit);
            st.setLong(2,minValue);
            st.setString(3,timeUnit);
            st.setLong(4,maxValue);
            rs = st.executeQuery();
            
            while (rs.next()) {
                // First column holds ID, we ignore that
                temp.add(rs.getInt(2));
            }

        } catch (SQLException e) {
            logger.error("Failed getting request type counts!",e);
        }
        cleanResultSet(rs);
        cleanStatement(st);

        // Dimension result array
        result = new int[temp.size()];

        // Tranfer values into it
        for (int i=0; i < temp.size(); i++) {
            result[i] = ((Integer) temp.get(i)).intValue();
        }
        return result;
    }

    
    private int[] getPercentages(int[] totals) {

        // Create storage array for period percentages by request type
        int[] percentages = new int[totals.length];
        
        // Calculate period grand total
        int sum = 0;
        for (int i=0; i < totals.length; i++) sum = sum + totals[i];
        
        // Calculate period percentages
        for (int i=0; i < totals.length; i++) {
            // Don't divide by zero, instead return -1.
            if (sum > 0) {
                percentages[i] = totals[i] * 100 / sum;
            } else {
                percentages[i] = -1;
            }
        }
        
        return percentages;
    }
    
    
    public String getActivitySummary() {

        StringBuffer result = new StringBuffer();

        // Get stat names
        Vector names = getRequestTypeNames();
        
        // Grab period subtotals by stat
        int[] h1Totals = getRequestTypeTotals("minute",0*60,1*60);
        int[] h2Totals = getRequestTypeTotals("minute",1*60,2*60);
        int[] h3Totals = getRequestTypeTotals("minute",2*60,3*60);
        
        int[] d1Totals = getRequestTypeTotals("hour",0*24,1*24);
        int[] d2Totals = getRequestTypeTotals("hour",1*24,2*24);
        int[] d3Totals = getRequestTypeTotals("hour",2*24,3*24);

        // Calculate period percentages
        int[] h1Percentages = getPercentages(h1Totals);
        int[] h2Percentages = getPercentages(h2Totals);
        int[] h3Percentages = getPercentages(h3Totals);
        
        int[] d1Percentages = getPercentages(d1Totals);
        int[] d2Percentages = getPercentages(d2Totals);
        int[] d3Percentages = getPercentages(d3Totals);

        // Format result table
        for (int i=0; i < names.size(); i++) {
            result.append("<tr><td class=name>");
            result.append(names.get(i));
            
            result.append("</td><td class=dark>");
            result.append(h1Totals[i]);
            result.append("</td><td class=dark>");
            result.append(h2Totals[i]);
            result.append("</td><td class=dark>");
            result.append(h3Totals[i]);
            
            result.append("</td><td class=light>");
            result.append(h1Percentages[i]);
            result.append("</td><td class=light>");
            result.append(h2Percentages[i]);
            result.append("</td><td class=light>");
            result.append(h3Percentages[i]);

            result.append("</td><td class=dark>");
            result.append(d1Totals[i]);
            result.append("</td><td class=dark>");
            result.append(d2Totals[i]);
            result.append("</td><td class=dark>");
            result.append(d3Totals[i]);

            result.append("</td><td class=light>");
            result.append(d1Percentages[i]);
            result.append("</td><td class=light>");
            result.append(d2Percentages[i]);
            result.append("</td><td class=light>");
            result.append(d3Percentages[i]);
            result.append("</td></tr>\n");
        }
        
        return result.toString();
    }
   
}

