package jwebcache;

import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;

import java.io.*;
import java.net.*;
import java.util.*;

import net.i2p.client.I2PClientFactory;
import net.i2p.client.streaming.I2PSocketManagerFactory;
import net.i2p.client.streaming.I2PSocketManager;
import net.i2p.client.streaming.I2PServerSocket;
import net.i2p.client.streaming.I2PSocket;
import net.i2p.data.Base32;
import net.i2p.data.Destination;
import net.i2p.I2PException;


public class JWebcache {

    // Our logger instance.
    private static Logger logger = Logger.getLogger(JWebcache.class);
    
    // Configuration.
    private Properties properties; // Our config properties.

    // I2P router access configuration.
    private String i2cpHost;
    private int i2cpPort;
    
    // I2P tunnel configuration.
    private String i2pInboundLength;
    private String i2pOutboundLength;
    private String i2pInboundLengthVariance;
    private String i2pOutboundLengthVariance;
    private String i2pInboundQuantity;
    private String i2pOutboundQuantity;
    private String i2pInboundBackupQuantity;
    private String i2pOutboundBackupQuantity;
    private String i2pInboundNickname;
    private String i2pOutboundNickname;
    
    // References to important objects.
    private Database database;
    private WorkerPool workerPool;
    private ResponseFormatter responseFormatter;
    
    // Our I2P communications classes.
    private Destination dest; // I2P destination loaded from keyfile.
    private I2PSocketManager socketManager; // I2P socket manager.
    private I2PServerSocket serverSocket; // I2P server socket.
    private I2PSocket socket; // Freshly accepted client connection socket.

    // Other variables.
    private final String KEYFILE_NAME = "jwebcache.key"; // Name of our I2P destination keyfile.

    // Stuff to make plugins easier
    private final File baseDir;
    private static JWebcache cache;

    /**
     *  JWebcache [-d baseDir] [start|stop]
     */
    public static void main(String[] args) {

        String dir = System.getProperty("user.dir");
        String command = "start";
        if (args.length <= 0) {
            // start in $CWD
        } else if (args[0].equals("-d") && args.length >= 2) {
            dir = args[1];
            if (args.length >= 3)
                command = args[2];
        } else {
            command = args[0];
        }

        // Configure logger, create main class.
        File logProps = new File(dir, "config/log4j.properties");
        PropertyConfigurator.configure(logProps.getAbsolutePath());

        if (command.equals("start")) {
            new JWebcache(dir);
        } else if (command.equals("stop")) {
            if (cache != null) {
                logger.info("Jwebcache stopping.");
                cache.shutdown();
                cache = null;
            } else {
                logger.info("Jwebcache not running.");
            }
        } else {
            throw new IllegalArgumentException("Unknown command: " + command);
        }
    }


    void shutdown() {
        (new shutdownThread()).run();
    }

    private class shutdownThread extends Thread {

        public void run() {

            logger.info("Shutdown thread running.");

            // If we have a server socket, close and null it.
            if (serverSocket != null) {
                try {
                    logger.info("Shutdown thread closing server socket.");
                    serverSocket.close();
                } catch (Exception e) {
                    logger.error("Failed closing server socket!",e);
                } finally {
                    serverSocket = null;
                }
            }

            // If we have a socket manager, destroy and null it.
            if (socketManager != null) {
                logger.info("Shutdown thread destroying socket manager.");
                socketManager.destroySocketManager();
                socketManager = null;
            }

            // Fire all worker threads.
            if (workerPool != null) {
                logger.info("Shutdown thread shutting down worker pool.");
                workerPool.shutdown();
                workerPool = null;
            }

            // If we have a database class, shutdown and null it.
            if (database != null) {
                logger.info("Shutdown thread shutting down database.");
                database.shutdown();
                database = null;
            }
            
            logger.info("Shutdown thread done.");
        }
    }


    private boolean createShutdownThread() {

        try {
            logger.info("Creating shutdown thread.");

            // Set up shutdown thread to perform cleanup
            Runtime.getRuntime().addShutdownHook(new shutdownThread());
        }
        catch (Exception e) {
            logger.fatal("Failed creating shutdown thread!",e);
            return false;
        }
        return true;
    }
        
    
    private boolean loadConfiguration() {
        
        FileInputStream fis = null;
        boolean result = false;
        
        try {
            logger.info("Loading configuration properties.");
            properties = new Properties();
            fis = new FileInputStream(new File(baseDir, "config/jwebcache.properties"));
            properties.load(fis);
            
            // Read configuration properties
            i2cpHost = properties.getProperty("JWebcache.i2cpHost","127.0.0.1");
            i2cpPort = Integer.parseInt(properties.getProperty("JWebcache.i2cpPort","7654"));
            i2pInboundLength = properties.getProperty("JWebcache.i2pInboundLength","3");
            i2pOutboundLength = properties.getProperty("JWebcache.i2pOutboundLength","3");
            i2pInboundLengthVariance = properties.getProperty("JWebcache.i2pInboundLengthVariance","0");
            i2pOutboundLengthVariance = properties.getProperty("JWebcache.i2pOutboundLengthVariance","0");
            i2pInboundQuantity = properties.getProperty("JWebcache.i2pInboundQuantity","2");
            i2pOutboundQuantity = properties.getProperty("JWebcache.i2pOutboundQuantity","2");
            i2pInboundBackupQuantity = properties.getProperty("JWebcache.i2pInboundBackupQuantity","0");
            i2pOutboundBackupQuantity = properties.getProperty("JWebcache.i2pOutboundBackupQuantity","0");
            i2pInboundNickname = properties.getProperty("JWebcache.i2pInboundNickname","jwebcache");
            i2pOutboundNickname = properties.getProperty("JWebcache.i2pOutboundNickname","jwebcache");
            
            // Success
            result = true;
            
        } catch (Exception e) {
            result = false;
            logger.fatal("Failed reading configuration properties!",e);
        }
        
        if (fis != null) {
            try {
                logger.debug("Closing configuration input stream.");
                fis.close();
            } catch (Exception e) {
                result = false;
                logger.fatal("Failed closing configuration input stream.",e);
            } finally {
                fis = null;
            }
        } else {
            result = false;
            logger.fatal("Configuration input stream already null.");
        }

        return result;
    }


    private boolean createDatabase() {

        if (!properties.contains("Database.location"))
            properties.setProperty("Database.location", baseDir.getAbsolutePath());
        try {
            logger.info("Creating database object.");
            database = new Database(properties);
            
            logger.info("Starting database maintenance thread.");
            database.setName("Database");
            database.start();
        }
        catch (Exception e) {
            // Don't show exception details: already shown upstream.
            logger.fatal("Failed creating database object.");
            return false;
        }
        return true;
    }
    
    
    private boolean createResponseFormatter() {
        if (!properties.contains("ResponseFormatter.location"))
            properties.setProperty("ResponseFormatter.location", baseDir.getAbsolutePath());
        try {
            logger.info("Creating response formatter object.");
            responseFormatter = new ResponseFormatter(database,properties);
        } catch (Exception e) {
            // Don't show exception details: already shown upstream.
            logger.fatal("Failed creating response formatter object.");
            return false;
        }
        return true;
    }
    
    
    private boolean createWorkerPool() {
        try {
            logger.info("Creating worker pool.");
            workerPool = new WorkerPool(database,responseFormatter,properties);
        } catch (Exception e) {
            // Don't show exception details: already shown upstream.
            logger.fatal("Failed creating worker pool!");
            return false;
        }
        return true;
    }

    
    private boolean checkOrGenerateKeyfile() {

        try {
            logger.info("Checking for keyfile.");

            // Create file object to access keyfile
            File keyFile = new File(baseDir, KEYFILE_NAME);

            // If keyfile missing
            if (!keyFile.exists()) {
                // Use client factory to create client
                // Use client to create destination into keyfile
                logger.info("Keyfile missing, generating new keyfile!");
                I2PClientFactory.createClient().createDestination(new FileOutputStream(keyFile));
            }
        }
        catch (Exception e) {
            logger.fatal("Failed checking or generating keyfile!",e);
            return false;
        }
        return true;
    }



    private boolean createI2PSocketManager() {

        try {
            logger.info("Creating socket manager from keyfile.");

            // Establish properties
            Properties p = new Properties();
            p.setProperty("inbound.length", i2pInboundLength);
            p.setProperty("outbound.length", i2pOutboundLength);
            p.setProperty("inbound.lengthVariance", i2pInboundLengthVariance);
            p.setProperty("outbound.lengthVariance", i2pOutboundLengthVariance);
            p.setProperty("inbound.quantity", i2pInboundQuantity);
            p.setProperty("outbound.quantity", i2pOutboundQuantity);
            p.setProperty("inbound.backupQuantity", i2pInboundBackupQuantity);
            p.setProperty("outbound.backupQuantity", i2pOutboundBackupQuantity);
            p.setProperty("inbound.nickname", i2pInboundNickname);
            p.setProperty("outbound.nickname", i2pOutboundNickname);

            // Use socket manager factory to create socket manager from keyfile
            File keyFile = new File(baseDir, KEYFILE_NAME);
            socketManager = I2PSocketManagerFactory.createManager(new FileInputStream(keyFile),i2cpHost,i2cpPort,p);

            if (socketManager == null) {
                logger.fatal("Got a null socketmanager, I2P router unreachable!");
                return false;
            }
        }
        catch (Exception e) {
            logger.fatal("Failed creating a socket manager!",e);
            return false;
        }
        return true;
    }

    

    private boolean printDestinationKey() {

        try {
            // Get destination from socket manager
            // Print our destkey
            dest = socketManager.getSession().getMyDestination();
            logger.info("Listening on destkey: " + dest.toBase64());
            logger.info("Server at: http://" + Base32.encode(dest.calculateHash().getData()) + ".b32.i2p/");
        }
        catch (Exception e) {
            logger.fatal("Failed getting our destination key.",e);
            return false;
        }
        return true;
    }


            
    private void closeSocket(I2PSocket socket) {
        if (socket != null) {
            try {
                logger.info("Closing socket.");
                // Make a small fuss, client was very quick indeed
                if (!socket.isClosed()) socket.close(); else logger.warn("Socket already closed.");
            } catch (IOException e) {
                logger.error("Failed closing socket.",e);
            } finally {
                socket = null;
            }
        } else logger.warn("Socket already null.");
    }



    public JWebcache() {
        this(System.getProperty("user.dir"));
    }

    /**
     *  Warning - doesn't return until done, stays in constructor forever
     */
    public JWebcache(String dir) {
        
        baseDir = new File(dir);

        logger.info("JWebcache 0.2.6 starting.");
        
        // Ensure that we have a shutdown handler
        if (!createShutdownThread()) throw new RuntimeException("shutdown");
        
        // Ensure we get a configuration loaded.
        if (!loadConfiguration()) throw new RuntimeException("load");
        
        // Create our database class and have it connect
        if (!createDatabase()) { shutdown(); throw new RuntimeException("db"); }
        
        // Create our response formatter and have it load templates
        if (!createResponseFormatter()) { shutdown(); throw new RuntimeException("response"); }
        
        // Load or generate a keyfile
        if (!checkOrGenerateKeyfile()) { shutdown(); throw new RuntimeException("key"); }

        // Create worker pool
        if (!createWorkerPool()) { shutdown(); throw new RuntimeException("pool"); }

        // now we are stoppable
        cache = this;

        // Create an I2PSocketManager from keyfile
        if (!createI2PSocketManager()) { shutdown(); throw new RuntimeException("socket"); }
        
        // Print destination key
        if (!printDestinationKey()) { shutdown(); throw new RuntimeException("print"); }

        // Get our server socket from the socket manager
        serverSocket = socketManager.getServerSocket();

        // Loop listening while the server socket is functional
        while (serverSocket != null) {

            // Use server socket to accept socket connection from client
            try { socket = serverSocket.accept(); }
             catch (SocketTimeoutException e) { logger.error("Accept caught SocketTimeoutException!",e); }
             catch (SocketException e) { logger.error("Accept caught SocketException!",e); }
             catch (I2PException e) { logger.error("Accept caught I2PException!",e); }

            // Take precautions against sockets being null
            // A null socket is returned when the server socket closes
            if (socket != null) {

                // Report client destkey
                String fromDestination = socket.getPeerDestination().toBase64();
                logger.info("Accepted connection from: " + fromDestination);

                logger.info("Generating a request ID.");
                long requestId = database.getRequestId();
                
                if (requestId < 0) {
                    logger.error("Failed generating request ID!");
                    logger.error("Closing socket, not attempting to handle!");
                    closeSocket(socket);
                    continue;
                }

                logger.info("Registering request " + requestId + " in database.");
                
                if (!database.registerRequest(requestId,fromDestination)) {
                    logger.error("Failed registering request " + requestId + " in database, closing socket!");
                    closeSocket(socket);
                    continue;
                }
                
                // Try assigning request to worker
                if (!workerPool.assignRequest(socket,requestId,fromDestination)) {
                    logger.error("Failed assigning request " + requestId + " to worker, closing socket!");
                    closeSocket(socket);
                    continue;
                }
            }
        }

        logger.info("Main class finished.");
    }

}









