package jwebcache;

import org.apache.log4j.Logger;

import java.util.*;
import java.io.*;


public class ResponseFormatter {
    
    // Our logger instance.
    private static Logger logger = Logger.getLogger(ResponseFormatter.class);
    
    // Configuration.
    private Properties properties; // Our config properties.
    private boolean hideErrorDetails; // Should we hide error details?
    
    // References to important objects.
    private Database database; // Database we use.

    // Template buffers.
    private String introTemplate;
    private String statsTemplate;
    private String errorTemplate;
    private String styleTemplate;
    
    // Template filenames.
    private final String INTRO_FILE = "config/introduction.html";
    private final String STATS_FILE = "config/statistics.html";
    private final String ERROR_FILE = "config/error.html";
    private final String STYLE_FILE = "config/style.css";

    // Replacement target markers.
    private final String STYLE_MARKER = "<!-- STYLE SHEET REPLACES THIS COMMENT -->";
    private final String ERROR_MARKER = "<!-- ERROR MESSAGE REPLACES THIS COMMENT -->";
    private final String PUT_QUOTA_MARKER = "<!-- PUT QUOTA STATS REPLACE THIS COMMENT -->";
    private final String TOTAL_QUOTA_MARKER = "<!-- TOTAL QUOTA STATS REPLACE THIS COMMENT -->";
    private final String ACTIVITY_MARKER = "<!-- ACTIVITY STATS REPLACE THIS COMMENT -->";
    private final String REQUESTS_MARKER = "<!-- RECENT REQUESTS REPLACE THIS COMMENT -->";
    private final String HEADERS_MARKER = "<!-- RECENT HEADERS REPLACE THIS COMMENT -->";
    

    public ResponseFormatter(Database database, Properties properties) throws Exception {
        
        // Grab arguments to local variables
        this.database = database;
        this.properties = properties;
        
        try {
            // Parse configuration properties
            hideErrorDetails = Boolean.parseBoolean(properties.getProperty("ResponseFormatter.hideErrorDetails","true"));

        } catch (Exception e) {
            logger.fatal("Failed reading config properties, response formatter constructor aborting.",e);
            throw new Exception("Response formatter constructor aborting.");
        }
        
        logger.info("Loading response templates.");
        
        // Load templates
        if (!loadTemplates()) {
            logger.fatal("Failed loading templates, response formatter constructor aborting.");
            throw new Exception("Response formatter constructor aborting.");
        }
    }
    
    
    public String formatIntroduction() {
        
        String response = introTemplate;
        
        // Package in "200 OK" header and return
        response = addHeader(response,"200 OK","text/html");
        return response;
    }
    
    
    public String formatStatistics() {
        
        String response = statsTemplate;
        
        // Splice in generated content
        response = response.replaceFirst(PUT_QUOTA_MARKER,database.getPutQuotaSummary());
        response = response.replaceFirst(TOTAL_QUOTA_MARKER,database.getTotalQuotaSummary());
        response = response.replaceFirst(ACTIVITY_MARKER,database.getActivitySummary());
        response = response.replaceFirst(REQUESTS_MARKER,database.getRecentRequests());
        response = response.replaceFirst(HEADERS_MARKER,database.getRecentHeaders());
            
        // Package in "200 OK" header and return
        response = addHeader(response,"200 OK","text/html");
        return response;
    }
    
    
    public String formatError(String errorMessage) {
        
        // Load resources
        String response = errorTemplate;
        
        // If showing error messages ain't permitted, replace error message with generic version
        if (hideErrorDetails) errorMessage = "An undisclosed error occurred while processing your request.";
            
        // Splice in error message
        response = response.replaceFirst(ERROR_MARKER,errorMessage);
            
        // Package in "403 Forbidden" header and return
        response = addHeader(response,"403 Forbidden","text/html");
        return response;
    }
    
    
    public String addHeader(String response, String errorCode, String contentType) {

        StringBuffer result = new StringBuffer("");
        
        result.append("HTTP/1.1 ");
        result.append(errorCode);
        result.append("\r\n");
        result.append("Connection: close\r\n");
        result.append("Content-Type: ");
        result.append(contentType);
        result.append("\r\n");
        result.append("Content-Length: ");
        result.append(response.length());
        result.append("\r\n\r\n");
        result.append(response);
        result.append("\r\n");

        return result.toString();
    }
    
    
    private boolean loadTemplates() {
        
        logger.info("Loading introduction template.");
        if ((introTemplate = loadFile(INTRO_FILE)) == null) return false;
        
        logger.info("Loading statistics template.");
        if ((statsTemplate = loadFile(STATS_FILE)) == null) return false;
        
        logger.info("Loading error template.");
        if ((errorTemplate = loadFile(ERROR_FILE)) == null) return false;
        
        logger.info("Loading style template.");
        if ((styleTemplate = loadFile(STYLE_FILE)) == null) return false;
        
        logger.info("Splicing style into introduction template.");
        introTemplate = introTemplate.replaceFirst(STYLE_MARKER,styleTemplate);
        
        logger.info("Splicing style into statistics template.");
        statsTemplate = statsTemplate.replaceFirst(STYLE_MARKER,styleTemplate);
        
        logger.info("Splicing style into error template.");
        errorTemplate = errorTemplate.replaceFirst(STYLE_MARKER,styleTemplate);
        
        return true;
    }
    

    private String loadFile(String filename) {
        
        File file;
        FileInputStream fis = null;
        String data = null;
        String location = properties.getProperty("ResponseFormatter.location");
        if (location == null)
            location = System.getProperty("user.dir");
        
        try {
            logger.debug("Opening file \"" + filename + "\".");
            file = new File(location, filename);
            fis = new FileInputStream(file);
            
            logger.debug("Loading file \"" + filename + "\" sized " + file.length() + " bytes.");
            byte buffer[] = new byte[ (int) file.length() ];
            fis.read(buffer);
            data = new String(buffer);

        } catch (Exception e) {
            logger.error("Failed loading file \"" + filename + "\".",e);
        }
        
        if (fis != null) {
            try {
                logger.debug("Closing file input stream.");
                fis.close();
            } catch (Exception e) {
                logger.error("Failed closing file input stream.",e);
            } finally {
                logger.debug("Nulling file input stream.");
                fis = null;
            }
        } else {
            logger.warn("File input stream already null.");
        }
        
        return data;
    }

}

